#include <QDebug>

#include "crfmodel.h"

//#define PRINT_NUMERICALLY_APPROX_SEC_ORDER_DERIVATIVES
//#define PRINT_NUMERICALLY_APPROX_DERIVATIVES
//#define PRINT_LEARNING_DEBUG_COMMENTS
//#define PRINT_MAP_DEBUG_COMMENTS

#define FeatureFunctionScaleDivision 10000 //Used to ensure arithmetic precision.

using namespace CRF;

CRFModel::CRFModel()
{
    imageWidth = 1;
    imageHeight = 1;
    MAPSteps = 10;

    //Initializing values for this particular model
    paramCount = 4;
    statesCount = 2;
}

CRFModel::CRFModel(int imgWidth, int imgHeight, int MAPSTEPS)
{
    imageWidth = imgWidth;
    imageHeight = imgHeight;
    MAPSteps = MAPSTEPS;

    //Initializing values for this particular model
    paramCount = 4;
    statesCount = 2;
}

int_int_vector CRFModel::EstimateMAP(const double_double_vector Input)
{

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    qDebug() << "Estimating MAP";
    #endif

    int_int_vector Ret(imageWidth);
    for (size_t i = 0; i < imageWidth; i++)
    {
        Ret[i] = int_vector(imageHeight);
        for (size_t j = 0; j < imageHeight; j++)
        {
            Ret[i][j] = Input[i][j];
        }
    }

    //Check if Input is correct format:
    if (Input.size() != imageWidth)
    {
        throw "Error! Input image doesn't have correct imageWidth.";
    }
    for (size_t i = 0; i < imageWidth; i++)
    {
        if (Input[i].size() != imageHeight)
        {
            throw "Error! Input image doesn't have correct imageHeight.";
        }
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    qDebug() << " Inputs are correct. Starting the loop for finding best solution...";
    #endif

    double BestStateEnergy = EvaluateOutputEnergy(Input, Ret);
    double LastBestStateEnergy = BestStateEnergy - 1;

    while (BestStateEnergy > LastBestStateEnergy)
    {
        LastBestStateEnergy = BestStateEnergy;
        for (size_t i=0; i < imageWidth; i++)
        {
            for (size_t j=0; j < imageHeight; j++)
            {
                size_t CurState = Ret[i][j];
                size_t BestState = CurState;
                for (size_t k = 0; k < statesCount; k++)
                {
                    double CurrentStateEnergy = EvaluateOutputEnergy(Input,  Ret);
                    if (k != CurState)
                    {
                        Ret[i][j] = k;
                        double NewStateEnergy = EvaluateOutputEnergy(Input,  Ret);

                        if (NewStateEnergy > CurrentStateEnergy)
                        {
                            #ifdef PRINT_MAP_DEBUG_COMMENTS
                            qDebug() << " Node(" <<  i << ", " << j << " changed state to:"  << k;
                            #endif

                            CurrentStateEnergy = NewStateEnergy;
                            BestState = k;
                        }
                    }
                }

                Ret[i][j] = BestState;
            }
        }

        BestStateEnergy = EvaluateOutputEnergy(Input, Ret);

        #ifdef PRINT_MAP_DEBUG_COMMENTS
        qDebug() << " New Iterated Conditional Modes Iteration: ";
        qDebug() << "   LastBestStateEnergy: " << LastBestStateEnergy;
        qDebug() << "   BestStateEnergy: " << BestStateEnergy;
        #endif
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    double eng = EvaluateOutputEnergy(Input, Ret);

    qDebug() << " Displaying Result...";
    qDebug() << " Output Energy: " << eng;
    for (size_t i = 0; i < imageWidth; i++)
    {
        for (size_t j = 0; j < imageHeight; j++)
        {
            qDebug() << " Input(" << i << ", " << j << ") = " << Input[i][j];
            qDebug() << " Result(" << i << ", " << j << ") = " << Ret[i][j];
        }
    }
    #endif

    return Ret;

}

double CRFModel::EvaluatePartitionFunction(const double_double_vector & Input)
{
    /* Slow implementation. Cycles through every single solution */
    size_t GridNodes = imageWidth * imageHeight;
    size_t NumberOfCombinations = pow(statesCount, GridNodes);

    size_t CurrentCombination = 0;

    double Ret = 0;

    while (CurrentCombination < NumberOfCombinations)
    {
        //Represent the current configuration
        int_vector T = ConvertToNewNumberRepresentation(CurrentCombination, statesCount, GridNodes);
        int_int_vector CurrentOutputVec(imageWidth);
        int c = 0;
        for (int a = 0; a < imageWidth; a++)
        {
            CurrentOutputVec[a] = int_vector(imageHeight);
            for (int b = 0; b < imageHeight; b++)
            {
                CurrentOutputVec[a][b] = T[c];
                c++;
            }
        }

        //Calculate the current configuration
        Ret += exp(EvaluateOutputEnergy(Input, CurrentOutputVec));

        CurrentCombination++;
    }

    return Ret;
}

double_vector CRFModel::EvaluateExpectedFeatureFunction(const double_double_vector & Input)
{
    size_t GridNodes = imageWidth * imageHeight;
    size_t NumberOfCombinations = pow(statesCount, GridNodes);

    size_t CurrentCombination = 0;

    double_vector Ret(paramCount + 1);
    while (CurrentCombination < NumberOfCombinations)
    {
        //Represent the current configuration as an int_int_vector
        int_vector T = ConvertToNewNumberRepresentation(CurrentCombination, statesCount, GridNodes);
        int_int_vector CurrentOutputVec(imageWidth);
        int c = 0;
        for (int a = 0; a < imageWidth; a++)
        {
            CurrentOutputVec[a] = int_vector(imageHeight);
            for (int b = 0; b < imageHeight; b++)
            {
                CurrentOutputVec[a][b] = T[c];
                c++;
            }
        }

        //Calculate on the current configuration
        double CurrentExpEnergy = exp(EvaluateOutputEnergy(Input, CurrentOutputVec));
        Ret[0] += CurrentExpEnergy; //<- Partition Function

        double_vector FeatureVectors(paramCount);
        for (size_t a = 0; a < imageWidth; a++)
        {
            for (size_t b = 0; b < imageHeight; b++)
            {
                double_vector NewVect = EvaluateNodeVector(Input, CurrentOutputVec, a, b);

                for (size_t k = 0; k < paramCount; k++)
                {
                    FeatureVectors[k] += NewVect[k];
                }
            }
        }

        //Calculate feature functions
        for (size_t r = 1; r < paramCount + 1; r++)
        {
                Ret[r] += FeatureVectors[r-1]*CurrentExpEnergy;
        }

        CurrentCombination++;
    }

    //Divide the feature functions with the partition function, so that they become expected values.
    for (size_t r = 1; r < paramCount + 1; r++)
    {
        Ret[r] = Ret[r]/Ret[0];
    }

    return Ret;
}

//Estimates the cost (likelihood) of the correct output given the corresponding input data
double CRFModel::EstimateLogLikelihood(const double_double_vector & Input, const int_int_vector & Output)
{
    return - log(EvaluatePartitionFunction(Input)) + EvaluateOutputEnergy(Input, Output);
}

//Estimates the model parameter by a set of corresponding input and output samples
void CRFModel::EstimateModelParameters(const std::vector<double_double_vector> & InputSamples, const std::vector<int_int_vector> & OutputSamples, size_t Steps, double StepFactor)
{
    #ifdef PRINT_LEARNING_DEBUG_COMMENTS
    qDebug() << "EstimateModelParameters";
    #endif

    weights = double_vector(paramCount);

    for (size_t k = 0; k < paramCount; k++)
    {
        weights[k] = 1;

    }

    //Maximum likelihood parameters (w.r.t. FeatureFunctionScaleDivision = 10000):
    /*
    weights[0] = -3136.12;
    weights[1] = 33192;
    weights[2] = 7909.15;
    weights[3] = -248.173;
    */

    //The log-likelihood of these is: -2.7376

    //Calculate fixed derivative term first, the emperical mean of the feature functions
    #ifdef PRINT_LEARNING_DEBUG_COMMENTS
    qDebug() << " Calculating first derivative term (the fixed part)...";
    #endif

    double_vector EmpericalMeanFeat(paramCount);

    for (size_t n = 0; n < InputSamples.size(); n++)
    {
        for (size_t a = 0; a < imageWidth; a++)
        {
            for (size_t b = 0; b < imageHeight; b++)
            {
                double_vector NewVect = EvaluateNodeVector(InputSamples[n], OutputSamples[n], a, b);

                for (size_t k = 0; k < paramCount; k++)
                {
                     EmpericalMeanFeat[k] += NewVect[k];
                }
            }
        }
    }

    double_vector FirstDerivativeTerm(paramCount);
    for (size_t k = 0; k < paramCount; k++)
    {
        FirstDerivativeTerm[k] = EmpericalMeanFeat[k];
    }

    #ifdef PRINT_LEARNING_DEBUG_COMMENTS
    qDebug() << " Calculating second derivative term (the almost intractable part)...";
    #endif

    for (size_t i = 0; i < Steps; i++)
    {
        double_vector SecondDerivativeTerm(paramCount);

        double NewLogLikelihood = 0;

        #ifdef PRINT_LEARNING_DEBUG_COMMENTS
        qDebug() << "   Iteration: " << i;
        #endif

        for (size_t n = 0; n < InputSamples.size(); n++)
        {
            #ifdef PRINT_LEARNING_DEBUG_COMMENTS
            //qDebug() << "    Sample: " << n;
            #endif

            double_vector TmpResult = EvaluateExpectedFeatureFunction(InputSamples[n]);

            for (size_t k = 1; k < paramCount + 1; k++)
            {
                SecondDerivativeTerm[k-1] += TmpResult[k];
            }

            double sampLglikelihood = EstimateLogLikelihood(InputSamples[n], OutputSamples[n]);
            #ifdef PRINT_LEARNING_DEBUG_COMMENTS
            //qDebug() << "    sampLglikelihood: " << sampLglikelihood;
            #endif

            NewLogLikelihood += sampLglikelihood;
        }

        //qDebug() << "    Finished calculating derivatives... updating weights.";

        for (size_t k = 0; k < paramCount; k++)
        {
            double drv = FirstDerivativeTerm[k] - SecondDerivativeTerm[k];

            weights[k] +=  FeatureFunctionScaleDivision*StepFactor*drv;

            #ifdef PRINT_LEARNING_DEBUG_COMMENTS
            qDebug() << "       New weight[" << k << "]: " << weights[k];
            #endif

            #ifdef PRINT_NUMERICALLY_APPROX_DERIVATIVES

            double delta = 0.00000000001;
            weights[k] = weights[k] + delta;

            double pr = 0;
            for (size_t n = 0; n < InputSamples.size(); n++)
            {
                pr += EstimateLogLikelihood(InputSamples[n], OutputSamples[n]);
            }

            weights[k] = weights[k] - 2*delta;

            double po = 0;
            for (size_t n = 0; n < InputSamples.size(); n++)
            {
                po += EstimateLogLikelihood(InputSamples[n], OutputSamples[n]);
            }

            weights[k] = weights[k] + delta;

            double DiffDeriv = fabs(drv - (pr - po)/(2*delta));

            qDebug() << "       Diif Derivative: " << QString::number(DiffDeriv, 'f', 10) ;
            qDebug() << "       Calculated Derivative: " << QString::number(drv, 'f', 10) ;
            qDebug() << "       Num. Approx. Derivative: " << QString::number((pr - po)/(2*delta), 'f', 10) ;

            //The following can be enabled to recover more information of iteration:
            //qDebug() << "  > FirstDerivativeTerm " << k << ": " << FirstDerivativeTerm[k];
            //qDebug() << "  > SecondDerivativeTerm " << k << ": " << SecondDerivativeTerm[k];

            #endif

            #ifdef PRINT_NUMERICALLY_APPROX_SEC_ORDER_DERIVATIVES
            double secdelta = 0.00000000001;
            double pSecondDerivativeTermDelta = 0;
            double qSecondDerivativeTermDelta = 0;

            weights[k] = weights[k] + secdelta;
            for (size_t n = 0; n < InputSamples.size(); n++)
            {
                pSecondDerivativeTermDelta -= EvaluateExpectedFeatureFunction(InputSamples[n])[k];
            }

            weights[k] = weights[k] - 2*secdelta;
            for (size_t n = 0; n < InputSamples.size(); n++)
            {
                qSecondDerivativeTermDelta -= EvaluateExpectedFeatureFunction(InputSamples[n])[k];
            }
            weights[k] =  weights[k] + secdelta;
            double approx = (pSecondDerivativeTermDelta - qSecondDerivativeTermDelta)/(2*secdelta);
            qDebug() << "       Second Order Derivative [ " << k << "]: " << QString::number(approx, 'f', 10) ;
            #endif

        }

        NewLogLikelihood = NewLogLikelihood / InputSamples.size();

        qDebug() << " Log-Likelihood: " << NewLogLikelihood; //Used for testing.

    }

    qDebug() << " The optimal parameters are:";
    for (int k = 0; k < paramCount; k++)
    {
        qDebug() << "  lambda[" << k << "]: " << weights[k] / FeatureFunctionScaleDivision;
    }
}

double CRFModel::EvaluateOutputEnergy(const double_double_vector & Input, const int_int_vector & Output)
{
    double R = 0;
    double_vector NodeVectorSum(paramCount);

    for (size_t i = 0; i < imageWidth; i++)
    {
        for (size_t j = 0; j < imageHeight; j++)
        {
            double_vector T = EvaluateNodeVector(Input, Output, i, j);
            for (size_t k = 0; k < paramCount; k++)
            {
                NodeVectorSum[k] += T[k];
            }
        }
    }

    for (size_t k = 0; k < paramCount; k++)
    {
        R += NodeVectorSum[k] * weights[k];
    }

    return R;
}

double CRFModel::EvaluateNodeEnergy(const double_double_vector Input, const int_int_vector Output, size_t i, size_t j)
{
    double Ret = 0;

    double_vector NodeVectorSum = EvaluateNodeVector(Input, Output, i, j);
    for (size_t k = 0; k < paramCount; k++)
    {
        Ret += NodeVectorSum[k] * weights[k];
    }

    return Ret;
}

double_vector CRFModel::EvaluateNodeVector(const double_double_vector Input, const int_int_vector Output, size_t i, size_t j)
{
    double_vector Ret(4);

    Ret[0] = Output[i][j];
    Ret[1] = Output[i][j]*Input[i][j];

    int mini=i-1;
    if (mini < 0)
    {
        mini = 0;
    }

    int minj=j-1;
    if (minj < 0)
    {
        minj = 0;
    }

    int maxi=i+1;
    if (maxi + 1 > imageWidth)
    {
        maxi = imageWidth-1;
    }

    int maxj=j+1;
    if (maxj + 1 > imageHeight)
    {
        maxj = imageHeight-1;
    }

    Ret[2] = 0;
    Ret[3] = 0;

    for (size_t newi=mini; newi <= maxi; newi++)
    {
        for (size_t newj=minj; newj <= maxj; newj++)
        {
            if (abs(Output[i][j] - Output[newi][newj]) > 0)
            {
                Ret[2] -= 1;
                Ret[3] -= exp(pow(Input[newi][newj] - Input[i][j], 2));
            }
        }
    }

    Ret[2] = Ret[2] / 8;
    Ret[3] = Ret[3] / 8;

    /*
    I need to rescale the feature functions here, so that the partition function
    (in later calculates) is sufficiently small. Otherwise the calculations will
    blow up when we divide with the partition function giving us an inaccurate result.
    */
    for (size_t k = 0; k < 4; k++)
    {
        Ret[k] = Ret[k]/FeatureFunctionScaleDivision;
    }

    return Ret;
}

int_vector CRFModel::ConvertToNewNumberRepresentation(int number, int newRepresentationSymbolCount, int maxDigits)
{
    int_vector Ret(maxDigits);
    int nnumb = number;
    int c = 0;

    while(nnumb > 0 && c < maxDigits)
    {
        int r = nnumb % newRepresentationSymbolCount ;
        Ret[c] = r ;
        nnumb = nnumb / newRepresentationSymbolCount;
        c++ ;
    }

    if (c > maxDigits)
    {
        throw "ConvertToNewNumberRepresentation tried to convert too big a number to a new numeral system.";
    }

    return Ret;
}

#undef FeatureFunctionScaleDivision

#undef PRINT_NUMERICALLY_APPROX_SEC_ORDER_DERIVATIVES
#undef PRINT_MAP_DEBUG_COMMENTS
#undef PRINT_LEARNING_DEBUG_COMMENTS
#undef PRINT_NUMERICALLY_APPROX_DERIVATIVES
