/**
 * @file crfmodel.h
 * @author  Iulian Vlad Serbane <julianserban@gmail.com>
 * @version 1.0
 * @section LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details at
 * http://www.gnu.org/copyleft/gpl.html
 *
 * @section DESCRIPTION
 *
 * The CRFModel_ExactLikelihood is a discrete image-denoising conditional random field model
 * using exact likelihood to calculate derivatives.
 *
 * @section DATAFORMAT
 *
 * All input images are expected to be of the format std::vector< std::vector<double> >,
 * where the first index indicates the horizontal location and the second the vertical location
 * of the pixel relative to origin (0,0) in the top left corner.
 * The values are doubles representing intensities of gray.
 *
 * All ouput images are expected to be of the format std::vector< std::vector<unsigned int> >,
 * where the first index indicates the horizontal location and the second the vertical location
 * of the pixel relative to origin (0,0) in the top left corner.
 * The values are int representing intensities of gray -
 * most often just {0,1} represnting black and white.
 */

#include <vector>
#include <stdlib.h>
#include <math.h>
#ifndef CRFMODEL_H
#define CRFMODEL_H

typedef std::vector<double> double_vector;
typedef std::vector< std::vector<double> > double_double_vector;

typedef std::vector<size_t> int_vector;
typedef std::vector< std::vector<size_t> > int_int_vector;

namespace CRF
{

    class CRFModel
    {
    public:
        /**
         * Constructor that initializes the configuration of the model. Width and Height are set to 1 pixel.
         */
        CRFModel();

        /**
         * Constructor that initializes the configuration of the model.
         * @param imgWidth Image width in pixels.
         * @param imgHeight Image height in pixels.
         * @param MAPSTEPS Maximum number of steps to perform MAP estimation.
         */
        CRFModel(int imgWidth, int imgHeight, int MAPSTEPS);

        /**
         * Estimates the MAP (Maximum Aposterori Estimation) solution. By default, the Iterated Conditional Modes algorithm is used.
         * @param Input Input sample.
         * @return MAP solution.
         */
        virtual int_int_vector EstimateMAP(const double_double_vector Input);

        /**
         * Estimates the exact loglikelihood for a given input sample and output.
         * @param Input Input sample.
         * @param Output Output sample.
         * @return MAP solution.
         */
        double EstimateLogLikelihood(const double_double_vector & Input, const int_int_vector & Output);

        /**
         * Estimates the model parameters using steepest decent.
         * @param InputSamples Array of input samples. The format is InputSamples[i] = Input Sample i.
         * @param OutputSamples Array of correctly classified output samples. The format is OutputSamples[i] = Output Sample i.
         * @param Steps Number of steps to perform steepest decent before terminating.
         * @param StepFactor Factor determining the step length (in the direction of the loglikelihood gradient) of each iteration. Should be greater than zero and likely less than one.
                             (The actual step length taken is multiplied by the preprocessor variable FeatureFunctionScaleDivision, which was divided by in the function EvaluateNodeVector to improve numerical accuracy on the partition function.
                             This is just a computational practicality and will not affect any other calculations.).
         */
        virtual void EstimateModelParameters(const std::vector<double_double_vector> & InputSamples, const std::vector<int_int_vector> & OutputSamples, size_t Steps, double StepFactor);

    protected:
        /**
         * The parameters of the model. The format is Weights[k] equals the weight of feature function k.
        */
        double_vector weights;

        /**
         * The image width. imageWidth times imageHeight equals the number nodes in the model.
         * For exact inference should be <= 4. Otherwise the calculations become intractable.
        */
        size_t imageWidth;

        /**
         * The image height. imageWidth times imageHeight equals the number nodes in the model.
         * For exact inference should be <= 4. Otherwise the calculations become intractable.
        */
        size_t imageHeight;

        /**
         * The number of weights (parameters). It has to equal the number of Weights (Weights.size()).
         * Standard value: 5. (This is model specific so changes here will require changes in the source code.)
        */
        size_t paramCount;

        /**
         * The number of states of each node.
         * Standard value: 2 for black and white. (This is model specific so changes here will require changes in the source code.)
        */
        size_t statesCount;

        /**
         * Number of maximum iterations for the Iterated Conditional Modes algorithm.
        */
        size_t MAPSteps;


        /**
         * Function used to estimate the exact partition function.
         *
         * @param Input Input sample.
         * @return The exact partition function.
         */
        double EvaluatePartitionFunction(const double_double_vector & Input);

        /**
         * Function used to estimate the exact partition function and the exact expected feature functions.
         *
         * @param Input Input sample.
         * @return The partition function and the expected feature functions (under the models pdf).
         *         The forrmat is: Ret[0] = Partition Function, Ret[1] = expected feature function 0, ..., Ret[paramCount + 1] =  value of expected feature function paramCount.
         */
        double_vector EvaluateExpectedFeatureFunction(const double_double_vector & Input);

        /**
         * Estimates the energy function for a given input and output sample.
         * @param Input Input sample.
         * @param Output Output sample.
         * @return The energy function.
        */
        double EvaluateOutputEnergy(const double_double_vector & Input, const int_int_vector & Output);

        /**
         * Estimates the energy at a specific node for a given input and output sample.
         * @param Input Input sample.
         * @param Output Output sample.
         * @return The energy for the specific node.
        */
        double EvaluateNodeEnergy(const double_double_vector Input, const int_int_vector Output, size_t i, size_t j);

        /**
         * Estimates the vector for a given node. Summing these vectors for all the nodes and taking the inner product with the Weights will yield the energy for a given input and output sample.
         * NOTICE: These values are divided by the preprocessor variable FeatureFunctionScaleDivision to improve numerical accuracy when doing calculations with the partition function.
         * @param Input Input sample.
         * @param Output Output sample.
         * @param i horizontal position of node.
         * @param j vertical position of node.
         * @return A node vector.
        */
        double_vector EvaluateNodeVector(const double_double_vector Input, const int_int_vector Output, size_t i, size_t j);

        /**
         * Converts an int (in the decimal numeral system, base-10) to an int [] (in any other numerical system, base-k).
         * This function is used by EvaluatePartitionFunction and EvaluateExpectedFeatureFunction to loop through all the possible configurations.
         * @param number Number to be converted.
         * @param newRepresentationSymbolCount Number of distinct symbols in the new numeral system.
         * @param maxDigits The maxmimum number of digits.
         * @return An int_vector where each entry in the array corresponds to a seperate symbol in the new numeral system. The vector has size maxDigits.
        */
        int_vector ConvertToNewNumberRepresentation(int number, int newRepresentationSymbolCount, int maxDigits);

    };


}

#endif // CRFMODEL_H
