#include <map>
#include <random>

#include <QDebug>

#include "crfmodel_cd.h"

using namespace CRF;

//#define PRINT_NUMERICALLY_APPROX_DERIVATIVES
//#define PRINT_LEARNING_DEBUG_COMMENTS //The CALCULATE_EXACT_LOGLIKELIHOOD should be enabled if the exact log-likelihood is needed.
//#define PRINT_MAP_DEBUG_COMMENTS
//#define CALCULATE_EXACT_LOGLIKELIHOOD
//#define CALCULATE_CD_BIAS

#define CD_USE_DATA_INITIALIZATION
//#define CD_USE_PERSISTENT_INITIALIZATION
//#define CD_USE_UNIFORM_INITIALIZATION

#define CD_INITIALIZE_PARAMETERS_RANDOMLY

#define FeatureFunctionScaleDivision 10000 //Used to ensure arithmetic precision.

#define Calculate_Statistics_Step 1 //The log-likelihood and bias components will be calculated at every CALCULATE_STATISTICS_STEP'th step.

CRFModel_CD::CRFModel_CD()
{
    imageWidth = 1;
    imageHeight = 1;
    MAPSteps = 10;

    //Initializing values for this particular model
    paramCount = 4;
    statesCount = 2;

    //In case this wasn't initialized already:
    chainSteps = 2;
    chainCount = 1;

    biasMax = double_vector(500);
    biasMin = double_vector(500);
    biasSum = double_vector(500);

    parameterNorm = double_double_vector(500);
    parameterSize = double_double_vector(500);

    logLikelihood = double_vector(500);

    testNumber = 0;
}

CRFModel_CD::CRFModel_CD(int imgWidth, int imgHeight, int MAPSTEPS)
{
    imageWidth = imgWidth;
    imageHeight = imgHeight;
    MAPSteps = MAPSTEPS;

    //Initializing values for this particular model
    paramCount = 4;
    statesCount = 2;

    //In case we didn't initialize this already:
    chainSteps = 2;
    chainCount = 1;

    biasMax = double_vector(500);
    biasMin = double_vector(500);
    biasSum = double_vector(500);

    parameterNorm = double_double_vector(500);
    logLikelihood = double_vector(500);

    parameterSize = double_double_vector(500);

    testNumber = 0;
}

void CRFModel_CD::SetCDParameters(int ChainSteps, int ChainCount)
{
    chainSteps = ChainSteps;
    chainCount = ChainCount;
}

void CRFModel_CD::EstimateModelParameters(const std::vector<double_double_vector> & InputSamples, const std::vector<int_int_vector> & OutputSamples, size_t Steps, double StepFactor)
{
    //Setup randomization for class, if the parameters are to be initialized at random values.
    #ifdef CD_INITIALIZE_PARAMETERS_RANDOMLY
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_real_distribution<> dis(0, 1);
    #endif

    parameterNorm[testNumber] = double_vector(Steps);

    //size_t chainSteps;
    //size_t chainCount;
    #ifdef PRINT_LEARNING_DEBUG_COMMENTS
    qDebug() << "EstimateModelParameters";
    #endif

    weights = double_vector(paramCount);
    parameterSize[testNumber] = double_vector(paramCount);

    for (size_t k = 0; k < paramCount; k++)
    {
        weights[k] = 1;

        //Initialize the parameters from a uniform distribution on [-5, 5].
        #ifdef CD_INITIALIZE_PARAMETERS_RANDOMLY
        std::random_device rd;
        std::mt19937 gen(rd());
        std::uniform_real_distribution<> dis(0, 1);

        weights[k] = 10*dis(gen) - 5;
        weights[k] = weights[k] * FeatureFunctionScaleDivision;
        #endif
    }

    //Set Persistent CD samples to the data. This is how Persisent CD is executed for the first learning iteration.
    #ifdef CD_USE_PERSISTENT_INITIALIZATION
    samplePCD = std::vector< std::vector<int_int_vector> >(InputSamples.size());
    for (size_t n = 0; n < InputSamples.size(); n++)
    {
        samplePCD[n] = std::vector<int_int_vector>(chainCount);
        for (int k = 0; k < chainCount; k++)
        {
            samplePCD[n][k] = int_int_vector(imageWidth);
            for (size_t a = 0; a < imageWidth; a++)
            {
                samplePCD[n][k][a] = int_vector(imageHeight);
                for (size_t b = 0; b < imageHeight; b++)
                {
                    samplePCD[n][k][a][b] = OutputSamples[n][a][b];
                }
            }
        }
    }
    #endif

    //Calculate fixed derivative term first, the emperical mean of the feature functions
    #ifdef PRINT_LEARNING_DEBUG_COMMENTS
    qDebug() << " Calculating first derivative term (the fixed part)...";
    #endif

    double_vector EmpericalMeanFeat(paramCount);

    for (size_t n = 0; n < InputSamples.size(); n++)
    {
        for (size_t a = 0; a < imageWidth; a++)
        {
            for (size_t b = 0; b < imageHeight; b++)
            {
                double_vector NewVect = EvaluateNodeVector(InputSamples[n], OutputSamples[n], a, b);

                for (size_t k = 0; k < paramCount; k++)
                {
                     EmpericalMeanFeat[k] += NewVect[k];
                }
            }
        }
    }

    double_vector FirstDerivativeTerm(paramCount);
    for (size_t k = 0; k < paramCount; k++)
    {
        FirstDerivativeTerm[k] = EmpericalMeanFeat[k];
    }

    #ifdef PRINT_LEARNING_DEBUG_COMMENTS
    qDebug() << " Calculating second derivative term (the almost intractable part)...";
    #endif

    int c = Calculate_Statistics_Step-1;

    for (size_t i = 0; i < Steps; i++)
    {
        c++;

        if (c > Calculate_Statistics_Step)
        {
            c = 1;
        }

        double_vector SecondDerivativeTerm(paramCount);

        #ifdef CALCULATE_CD_BIAS
        double_vector ExactSecondDerivativeTerm(paramCount);
        #endif

        double NewLogLikelihood = 0;

        #ifdef PRINT_LEARNING_DEBUG_COMMENTS
        qDebug() << "   Iteration: " << i;
        #endif


        for (size_t n = 0; n < InputSamples.size(); n++)
        {
            #ifdef PRINT_LEARNING_DEBUG_COMMENTS
            //qDebug() << "    Sample: " << n;
            #endif

            double_vector TmpResult = ApproximateExpectedFeatureFunctions(InputSamples[n], OutputSamples[n], n);

            for (size_t k = 1; k < paramCount + 1; k++)
            {
                SecondDerivativeTerm[k-1] += TmpResult[k];
            }


            double sampLglikelihood = 0;

            #ifdef CALCULATE_EXACT_LOGLIKELIHOOD
            if (c == Calculate_Statistics_Step)
            {
                sampLglikelihood = EstimateLogLikelihood(InputSamples[n], OutputSamples[n]);

                #ifdef CALCULATE_CD_BIAS
                double_vector ExactTmpResult = EvaluateExpectedFeatureFunction(InputSamples[n]);

                for (size_t k = 1; k < paramCount + 1; k++)
                {
                    ExactSecondDerivativeTerm[k-1] += ExactTmpResult[k];
                }
                #endif

            }


            #endif

            NewLogLikelihood += sampLglikelihood;
        }

        #ifdef PRINT_LEARNING_DEBUG_COMMENTS
        //qDebug() << "    Finished calculating derivatives... updating weights.";
        #endif

        double BiggestBias = 0;
        double SmallestBias = 0;

        parameterNorm[testNumber][i] = 0;

        for (size_t k = 0; k < paramCount; k++)
        {
            double drv = FirstDerivativeTerm[k] - SecondDerivativeTerm[k];
            weights[k] +=  FeatureFunctionScaleDivision*StepFactor*drv;

            #ifdef PRINT_LEARNING_DEBUG_COMMENTS
            qDebug() << "       New Weight " << k << ": " << weights[k];
            #endif

            #ifdef CALCULATE_CD_BIAS
            if (c == CALCULATE_STATISTICS_STEP)
            {
                double exactdrv = FirstDerivativeTerm[k] - ExactSecondDerivativeTerm[k];

                if (BiggestBias < fabs(drv - exactdrv) || k == 0)
                {
                    BiggestBias = fabs(drv - exactdrv);
                }
                if (SmallestBias > fabs(drv - exactdrv) || k == 0)
                {
                    SmallestBias = fabs(drv - exactdrv);
                }

                biasSum[i] += fabs(drv - exactdrv);
            }
            #endif

            parameterNorm[testNumber][i] += pow(weights[k] / FeatureFunctionScaleDivision, 2);
        }

        NewLogLikelihood = NewLogLikelihood / InputSamples.size();

        qDebug() << " Iteration " << i;

        #ifdef CALCULATE_EXACT_LOGLIKELIHOOD
        if (c == Calculate_Statistics_Step)
        {
            logLikelihood[i] = NewLogLikelihood;
            biasMax[i] = BiggestBias;
            biasMin[i] = SmallestBias;
        }

        qDebug() << "   LogLikelihood " << NewLogLikelihood; //Used for testing.

        #endif

    }

    //Used for testing:
    /*for (int k = 0; k < paramCount; k++)
    {
        //qDebug() << " weights[" << k << "]: " << weights[k];
    }*/

    #ifdef CALCULATE_EXACT_LOGLIKELIHOOD
    c = Calculate_Statistics_Step-1;
    for (int i = 0; i < Steps; i++)
    {
        c++;
        if (c > Calculate_Statistics_Step)
        {
            c = 1;
        }

        if (c == Calculate_Statistics_Step)
        {
           // qDebug() << " Statistics[" << i << "]: " << logLikelihood[i] << biasMin[i] << biasSum[i] << biasMax[i];  //Used for experiments.
        }
    }
    #endif

    for (int k = 0; k < paramCount; k++)
    {
        parameterSize[testNumber][k] = weights[k] / FeatureFunctionScaleDivision;
    }

    double_vector parameterNorm_Average(Steps);
    for (int i = 0; i < Steps; i++)
    {
        parameterNorm_Average[i] = 0;
        for (int k = 0; k <= testNumber; k++)
        {
            parameterNorm_Average[i] += parameterNorm[k][i];
        }
        parameterNorm_Average[i] = parameterNorm_Average[i] / (testNumber + 1);
    }

    double_vector parameterNorm_Variance(Steps);
    for (int i = 0; i < Steps; i++)
    {
        parameterNorm_Variance[i] = 0;
        for (int k = 0; k <= testNumber; k++)
        {
            parameterNorm_Variance[i] += pow(parameterNorm_Average[i] - parameterNorm[k][i], 2);
        }
        parameterNorm_Variance[i] = parameterNorm_Variance[i] / (testNumber + 1);
    }

    //Used for experiments:
    /*
    for (int i = 0; i < Steps; i++)
    {
        qDebug() << " parameterNorm[" << i << "]: " << parameterNorm_Average[i]; //<< parameterNorm_Variance[i] << parameterNorm_Min[i] << parameterNorm_Max[i];
    }
    */

    //Used for experiments:
    /*
    //Output all parameter values for the last learning iteration, for the last testNumber tests. This will be used for histogram plots.
    qDebug() << "#Output Parameters";
    for (int k = 0; k <= testNumber; k++)
    {
        qDebug() << parameterSize[k][0] << parameterSize[k][1]  << parameterSize[k][2]  << parameterSize[k][3];
    }
    qDebug() << "#End Output";
    */

    qDebug() << " The optimal parameters are:";
    for (int k = 0; k < paramCount; k++)
    {
        qDebug() << "  lambda[" << k << "]: " << weights[k] / FeatureFunctionScaleDivision;
    }

    testNumber++;
}

double_vector CRFModel_CD::ApproximateExpectedFeatureFunctions(const double_double_vector & Input, const int_int_vector & Output, size_t SampleIndex)
{
    double_vector Ret(paramCount + 1);

    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_real_distribution<> dis(0, 1);

    for (size_t r = 0; r < chainCount; r++)
    {
        //Setup chain...

        //This will initialize the Gibbs chain according to the last sample
        #ifdef CD_USE_PERSISTENT_INITIALIZATION
        int_int_vector ChainOutput(imageWidth);
        for (size_t a = 0; a < imageWidth; a++)
        {
            ChainOutput[a] = int_vector(imageHeight);
            for (size_t b = 0; b < imageHeight; b++)
            {
                ChainOutput[a][b] = samplePCD[SampleIndex][r][a][b];
            }
        }
        #endif

        //This will initialize the Gibbs chain according to the data
        #ifdef CD_USE_DATA_INITIALIZATION
        int_int_vector ChainOutput(imageWidth);
        for (size_t a = 0; a < imageWidth; a++)
        {
            ChainOutput[a] = int_vector(imageHeight);
            for (size_t b = 0; b < imageHeight; b++)
            {
                ChainOutput[a][b] = Output[a][b];
            }
        }
        #endif

        //This will initialize the Gibbs chain according to a uniform distribution.
        #ifdef CD_USE_UNIFORM_INITIALIZATION
        int_int_vector ChainOutput(imageWidth);
        for (size_t a = 0; a < imageWidth; a++)
        {
            ChainOutput[a] = int_vector(imageHeight);
            for (size_t b = 0; b < imageHeight; b++)
            {
                double RandValue = dis(gen)*statesCount;

                size_t NewState = statesCount - 1;

                double totEnergySoFar = 0;
                for (size_t d = 0; d < statesCount - 1; d++)
                {
                    totEnergySoFar++;
                    if (RandValue < totEnergySoFar)
                    {
                        NewState = d;
                        d = statesCount; //Break the loop. We found a new state.
                    }
                }

                ChainOutput[a][b] = NewState;
            }
        }
        #endif

        //Run chain...
        for (size_t k = 0; k < chainSteps; k++)
        {
            for (size_t a = 0; a < imageWidth; a++)
            {
                for (size_t b = 0; b < imageHeight; b++)
                {
                    /*
                    Loop through every state and calculate it's energy and the total energy of all states.
                    We are calculating the CRF model conditioned on all the other states.
                    This boils down to calculating a miniature CRF model with a single hidden node.
                    Here the total energy equals the partition function.
                    */

                    double_vector EnergyState(statesCount);
                    double totEnergy = 0;
                    for (size_t d = 0; d < statesCount; d++)
                    {
                        ChainOutput[a][b] = d;
                        EnergyState[d] = exp(EvaluateNodeEnergy(Input, ChainOutput, a, b));
                        totEnergy += EnergyState[d];
                    }


                    /*
                    We don't want to risk numerical inaccuracy by dividing a small number with huge one.
                    So instead of dividing with the total energy we draw a random number uniformly from the interval [0, total energy]
                    and use this for picking our a new state.
                    */
                    double RandValue = dis(gen)*totEnergy;

                    size_t NewState = statesCount - 1;

                    double totEnergySoFar = 0;
                    for (size_t d = 0; d < statesCount - 1; d++)
                    {
                        ChainOutput[a][b] = d;
                        totEnergySoFar += EnergyState[d];

                        if (RandValue < totEnergySoFar)
                        {
                            NewState = d;
                            d = statesCount; //Break the loop. We found a new state.
                        }
                    }

                    ChainOutput[a][b] = NewState;
                }
            }
        }

        //Calculate on chain
        //qDebug() << "  Calculate on chain...";
        double CurrentExpEnergy = exp(EvaluateOutputEnergy(Input, ChainOutput));
        Ret[0] += CurrentExpEnergy; //<- "Partition" Function

        double_vector FeatureVectors(paramCount);
        for (size_t a = 0; a < imageWidth; a++)
        {
            for (size_t b = 0; b < imageHeight; b++)
            {
                double_vector NewVect = EvaluateNodeVector(Input, ChainOutput, a, b);

                for (size_t k = 0; k < paramCount; k++)
                {
                    FeatureVectors[k] += NewVect[k];
                }
            }
        }

        //Calculate feature functions
        for (size_t r = 1; r < paramCount + 1; r++)
        {
            Ret[r] += FeatureVectors[r-1]*CurrentExpEnergy;
        }

        //If using Persistent CD, update the last sample to the current sample.
        #ifdef CD_USE_PERSISTENT_INITIALIZATION
        for (size_t a = 0; a < imageWidth; a++)
        {
            for (size_t b = 0; b < imageHeight; b++)
            {
                samplePCD[SampleIndex][r][a][b] = ChainOutput[a][b];
            }
        }
        #endif
    }

    //Divide the feature functions with the "partition" function, so that they become approximated expected values
    for (size_t r = 1; r < paramCount + 1; r++)
    {
        Ret[r] = Ret[r]/Ret[0];
    }

    return Ret;
}

#undef FeatureFunctionScaleDivision

#undef CD_INITIALIZE_PARAMETERS_RANDOMLY

#undef CD_USE_DATA_INITIALIZATION
#undef CD_USE_PERSISTENT_INITIALIZATION
#undef CD_USE_UNIFORM_INITIALIZATION

#undef CALCULATE_EXACT_LOGLIKELIHOOD

#undef PRINT_MAP_DEBUG_COMMENTS
#undef PRINT_LEARNING_DEBUG_COMMENTS
#undef PRINT_NUMERICALLY_APPROX_DERIVATIVES

