/**
 * @file crfmodel_cd.h
 * @author  Iulian Vlad Serbane <julianserban@gmail.com>
 * @version 1.0
 * @section LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details at
 * http://www.gnu.org/copyleft/gpl.html
 */

#ifndef CRFMODEL_CD_H
#define CRFMODEL_CD_H

#include <crfmodel_graphcut.h>

namespace CRF
{
    class CRFModel_CD : public CRFModel_GraphCut
    {
    public:
        /**
         * Constructor that initializes the configuration of the model. Width and Height are set to 1 pixel.
         */
        CRFModel_CD();

        /**
         * Constructor that initializes the configuration of the model.
         * @param imgWidth Image width in pixels.
         * @param imgHeight Image height in pixels.
         * @param MAPSTEPS Maximum number of steps to perform MAP estimation.
         */
        CRFModel_CD(int imgWidth, int imgHeight, int MAPSTEPS);

        /**
         * Sets the parameters of the contrastive divergence learning / inference methods.
         * @param ChainSteps Number of iterations each chain takes.
         * @param ChainCount Number of chains to average over.
         */
        void SetCDParameters(int ChainSteps, int ChainCount);

        /**
         * Estimates the model parameters using contrastive divergence.
         * @param InputSamples Array of input samples. The format is InputSamples[i] = Input Sample i.
         * @param OutputSamples Array of correctly classified output samples. The format is OutputSamples[i] = Output Sample i.
         * @param Steps Number of steps to perform steepest decent before terminating.
         * @param StepFactor Factor determining the step length (in the direction of the loglikelihood gradient) of each iteration. Should be greater than zero and likely less than one.
                             (The actual step length taken is multiplied by the preprocessor variable FeatureFunctionScaleDivision, which was divided by in the function EvaluateNodeVector to improve numerical accuracy on the partition function.
                             This is just a computational practicality and will not affect any other calculations.).
         */
        void EstimateModelParameters(const std::vector<double_double_vector> & InputSamples, const std::vector<int_int_vector> & OutputSamples, size_t Steps, double StepFactor);

        /**
         * Approximates the expected feature functions using Gibbs Chains.
         *
         * @param Input Input sample.
         * @param Output Correct output sample used to initialize the Gibbs chain.
         * @param SampleIndex The current training sample index.
         * @return A pseudo partition function (which is not to be used for further calculations) and the feature functions (under the models pdf).
         *         The forrmat is: Ret[0] = Partition Function, Ret[1] = expected feature function 0, ..., Ret[paramCount + 1] =  value of expected feature function paramCount.
         */
        double_vector ApproximateExpectedFeatureFunctions(const double_double_vector & Input, const int_int_vector & Output, size_t SampleIndex);

        /**
         * Number of iterations each chain takes.
        */
        size_t chainSteps;
        /**
         * Number of chains to average over.
        */
        size_t chainCount;

        /**
         * The current test number. Used for experiments.
        */
        size_t testNumber;
    private:
        /**
         * The sum of absolute biases for each iteration. Used for experiments.
        */
        double_vector biasSum;

        /**
         * The sum of absolute max bias for each iteration. Used for experiments.
        */
        double_vector biasMax;

        /**
         * The sum of absolute biases for each iteration. Used for experiments.
        */
        double_vector biasMin;

        /**
         * The sum of log-likelihood for each iteration. Used for experiments.
        */
        double_vector logLikelihood;

        /**
         * The sum of the parameter norm for each iteration. Used for experiments.
        */
        double_double_vector parameterNorm;

        /**
         * The sum of the parameter norm for each iteration. Used for experiments.
        */
        double_double_vector parameterSize;

        /**
         * The last Gibbs sample sample, used by the Persistent CD for initialization.
           The format is: samplePCD[SampleIndex][ChainNumber].
        */
        std::vector< std::vector<int_int_vector> > samplePCD;



    };
}
#endif // CRFMODEL_CD_H
