#include "crfmodel_graphcut.h"

using namespace CRF;

//#define PRINT_NUMERICALLY_APPROX_DERIVATIVES
//#define PRINT_LEARNING_DEBUG_COMMENTS //The CALCULATE_EXACT_LOGLIKELIHOOD should be enabled if the exact loglikelihood is needed.
//#define PRINT_MAP_DEBUG_COMMENTS
//#define CALCULATE_EXACT_LOGLIKELIHOOD

#define FeatureFunctionScaleDivision 10000 //Used to ensure arithmetic precision.

CRFModel_GraphCut::CRFModel_GraphCut()
{
    imageWidth = 1;
    imageHeight = 1;
    MAPSteps = 10;

    //Initializing values for this particular model
    paramCount = 4;
    statesCount = 2;
}

CRFModel_GraphCut::CRFModel_GraphCut(int imgWidth, int imgHeight, int MAPSTEPS)
{
    imageWidth = imgWidth;
    imageHeight = imgHeight;
    MAPSteps = MAPSTEPS;

    //Initializing values for this particular model (and Graph CUts can maximum optimize exactly for two states)
    paramCount = 4;
    statesCount = 2;
}

int_int_vector CRFModel_GraphCut::EstimateMAP(const double_double_vector Input)
{
    //Parameters produced by CD-20 after 100 iterations:
    /*
    weights[0] = -1886;
    weights[1] = 17538;
    weights[2] = 3129;
    weights[3] = 0;
    */

    double_vector adjustedWeights(paramCount);
    for (size_t k = 0; k < paramCount; k++)
    {
        adjustedWeights[k] = weights[k];
    }

    /*If the model was fitted outside the parameter space, where
    Graph Cuts are applicable, alter the parameters so that
    one can still apply Graph Cuts*/

    if (adjustedWeights[2] < 0)
    {
        adjustedWeights[2] = 0;
    }
    if (adjustedWeights[3] < 0)
    {
        adjustedWeights[3] = 0;
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    qDebug() << "Estimating MAP (Graph Cut Method)";
    for (size_t k = 0; k < paramCount; k++)
    {
        qDebug() << "adjustedWeights[" << k << "]: " << adjustedWeights[k];
    }
    #endif

    int_int_vector Ret(imageWidth);
    for (size_t i = 0; i < imageWidth; i++)
    {
        Ret[i] = int_vector(imageHeight);
        for (size_t j = 0; j < imageHeight; j++)
        {
            Ret[i][j] = 1; //Input[i][j];
        }
    }

    //Check if Input is correct format:
    if (Input.size() != imageWidth)
    {
        throw "Error! Input image doesn't have correct imageWidth.";
    }
    for (size_t i = 0; i < imageWidth; i++)
    {
        if (Input[i].size() != imageHeight)
        {
            throw "Error! Input image doesn't have correct imageHeight.";
        }
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    qDebug() << " Inputs are correct... Starting Graph Cut Procedure.";
    #endif

    int imageNodes = imageWidth*imageHeight;
    //map from vertices one-dimensional array -> image two-dimensional array
    int_vector MapVertexToX(imageNodes);
    int_vector MapVertexToY(imageNodes);

    size_t a = 0;
    size_t b = 0;

    for (int i = 0; i < imageNodes; i++)
    {
        //Create map from vertices one-dimensional array -> image two-dimensional array
        MapVertexToX[i] = a;
        MapVertexToY[i] = b;

        b++;
        if (b > imageHeight - 1)
        {
            b = 0;
            a++;
        }
    }


    typedef Graph<double,double,double> GraphType;

    int maxEdges = imageNodes*8 + imageNodes*2;

    GraphType G(static_cast<int>(imageNodes), static_cast<int>(maxEdges));

    G.add_node(imageNodes);

    double_vector AdditionalSourceWeight(imageNodes);
    double_vector AdditionalSinkWeight(imageNodes);
    for(int node = 0; node < imageNodes; node++)
    {
        for(int neighbour_node = 0; neighbour_node < imageNodes; neighbour_node++)
        {
            if (node < neighbour_node)
            {
                if (abs(MapVertexToX[node]-MapVertexToX[neighbour_node]) < 2 && abs(MapVertexToY[node]-MapVertexToY[neighbour_node]) < 2)
                {
                    double diff = Input[MapVertexToX[node]][MapVertexToY[node]] - Input[MapVertexToX[neighbour_node]][MapVertexToY[neighbour_node]];
                    double v = (adjustedWeights[2] + adjustedWeights[3]*exp(pow(diff, 2)))/8;

                    //connection node to neighbour_node with weight rsb
                    G.add_edge(node, neighbour_node, 4*v, 0);

                    #ifdef PRINT_MAP_DEBUG_COMMENTS
                    qDebug() << "cap: " << node << " -> " << neighbour_node << ": " << 4*v;
                    #endif

                    AdditionalSourceWeight[node] += 2*v;
                    AdditionalSinkWeight[neighbour_node] += 2*v;
                }
            }
        }
    }

    for(int node = 0; node < imageNodes; node++)
    {
        double cap = adjustedWeights[0] + adjustedWeights[1]*Input[MapVertexToX[node]][MapVertexToY[node]] - AdditionalSourceWeight[node] + AdditionalSinkWeight[node];

        if (cap < 0)
        {
            G.add_tweights(node,-cap,0);
            #ifdef PRINT_MAP_DEBUG_COMMENTS
            qDebug() << "cap: source -> " << node << ": " << -cap ;
            #endif
        }
        else
        {
            G.add_tweights(node,0,cap);
            #ifdef PRINT_MAP_DEBUG_COMMENTS
            qDebug() << "cap: " << node << " -> sink: " << cap ;
            #endif
        }
    }

    double flow = G.maxflow();

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    qDebug() << "flow: " << flow;
    #endif

    for(int node = 0 ; node < imageNodes; node++)
    {
        if(G.what_segment(node)== GraphType::SINK)
        {
            Ret[MapVertexToX[node]][MapVertexToY[node]] = 1;
        }
        else
        {
            Ret[MapVertexToX[node]][MapVertexToY[node]] = 0;
        }
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    double eng = EvaluateOutputEnergy(Input, Ret);
    qDebug() << " Output Energy: " << eng;
    #endif

    return Ret;
}



#undef FeatureFunctionScaleDivision

#undef CALCULATE_EXACT_LOGLIKELIHOOD

#undef PRINT_MAP_DEBUG_COMMENTS
#undef PRINT_LEARNING_DEBUG_COMMENTS
#undef PRINT_NUMERICALLY_APPROX_DERIVATIVES

