int_int_vector CRFModel_GraphCut::EstimateMAP(const double_double_vector Input)
{
    //Optimal parameters:
    weights[0] = -1886;
    weights[1] = 17538;
    weights[2] = 3129;
    weights[3] = 0;

    double_vector adjustedWeights(paramCount);
    for (int k = 0; k < paramCount; k++)
    {
        adjustedWeights[k] = weights[k];
    }

    /*If the model was fitted outside the parameter space where
    Graph Cuts are applicable, alter the paramters so that
    we can still apply Graph Cuts*/

    if (adjustedWeights[2] < 0)
    {
        adjustedWeights[2] = 0;
    }
    if (adjustedWeights[3] < 0)
    {
        adjustedWeights[3] = 0;
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    qDebug() << "Estimating MAP (Graph Cut Method)";
    for (int k = 0; k < paramCount; k++)
    {
        qDebug() << "adjustedWeights[" << k << "]: " << adjustedWeights[k];
    }
    #endif

    int_int_vector Ret(imageWidth);
    for (size_t i = 0; i < imageWidth; i++)
    {
        Ret[i] = int_vector(imageHeight);
        for (size_t j = 0; j < imageHeight; j++)
        {
            Ret[i][j] = 0; //Input[i][j];
        }
    }

    //Check if Input is correct format:
    if (Input.size() != imageWidth)
    {
        throw "Error! Input image doesn't have correct imageWidth.";
    }
    for (size_t i = 0; i < imageWidth; i++)
    {
        if (Input[i].size() != imageHeight)
        {
            throw "Error! Input image doesn't have correct imageHeight.";
        }
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    qDebug() << " Inputs are correct... Starting Graph Cut Procedure.";
    #endif

    //Build Graph

    using namespace boost;

    typedef adjacency_list_traits<vecS, vecS, directedS> Traits;
    typedef adjacency_list<vecS, vecS, directedS,
      property<vertex_name_t, std::string>,
      property<edge_capacity_t, long,
        property<edge_residual_capacity_t, long,
          property<edge_reverse_t, Traits::edge_descriptor> > > > Graph;

    typedef typename graph_traits<Graph>::vertices_size_type vertices_size_type;
    typedef typename graph_traits<Graph>::vertex_descriptor vertex_descriptor;
    typedef typename graph_traits<Graph>::edge_descriptor edge_descriptor;

    std::vector<vertex_descriptor> verts;

    Graph g;

    property_map<Graph, edge_capacity_t>::type
      capacity = get(edge_capacity, g);
    property_map<Graph, edge_reverse_t>::type
      reverse_edge = get(edge_reverse, g);
    property_map<Graph, edge_residual_capacity_t>::type
      residual_capacity = get(edge_residual_capacity, g);

    //Traits::vertex_descriptor s, t;

    //Build flow network

    int verticesCount = 2 + imageWidth*imageHeight;

    //map from vertices one-dimensional array -> image two-dimensional array
    int_vector MapVertexToX(verticesCount);
    int_vector MapVertexToY(verticesCount);

    int a = 0;
    int b = 0;

    for (int i = 0; i < verticesCount; i++)
    {
        verts.push_back(add_vertex(g));

        if (i > 0 && i < verticesCount - 1)
        {
            //Create map from vertices one-dimensional array -> image two-dimensional array
            MapVertexToX[i-1] = b;
            MapVertexToY[i-1] = a;

            //qDebug() << "Mapping " << i - 1 << " to " << MapVertexToX[i-1] << ", " << MapVertexToY[i-1];

            a++;
            if (a > imageWidth - 1)
            {
                a = 0;
                b++;
            }
        }
    }

    typename graph_traits<Graph>::vertex_descriptor& src = verts[0];
    typename graph_traits<Graph>::vertex_descriptor& sink = verts[verticesCount - 1];

    /*Establish capacities according to graph cut method... */
    int i = 1;
    for (int a = 0; a < imageWidth; a++)
    {
        for (int b = 0; b < imageHeight; b++)
        {
            //Add edges between terminal nodes and node [a][b]
            edge_descriptor e1, e2, e3, e4;
            bool in1, in2, in3, in4;

            boost::tie(e1, in1) = add_edge(verts[0], verts[i], g);
            boost::tie(e2, in2) = add_edge(verts[i], verts[0], g);
            capacity[e1] = 0;

            if (adjustedWeights[0] < 0)
            {
                capacity[e1] = capacity[e1] - (long)adjustedWeights[0];
            }

            if (adjustedWeights[1]*Input[a][b] < 0)
            {
                capacity[e1] = capacity[e1] - (long)adjustedWeights[1]*Input[a][b];
            }

            capacity[e2] = 0;  //Always zero by construction
            reverse_edge[e1] = e2;
            reverse_edge[e2] = e1;

            boost::tie(e3, in3) = add_edge(verts[verticesCount-1], verts[i], g);
            boost::tie(e4, in4) = add_edge(verts[i], verts[verticesCount-1], g);
            capacity[e3] = 0; //Always zero by construction

            capacity[e4] = 0;
            if (adjustedWeights[0] > 0)
            {
                capacity[e4] = capacity[e4] + (long)adjustedWeights[0];
            }

            if (adjustedWeights[1]*Input[a][b] > 0)
            {
                capacity[e4] = capacity[e4] + (long)adjustedWeights[1]*Input[a][b];
            }

            reverse_edge[e3] = e4;
            reverse_edge[e4] = e3;

            i++;
        }
    }

    i = 1;
    for (int a = 0; a < imageWidth; a++)
    {
        for (int b = 0; b < imageHeight; b++)
        {
            int j = 1;
            for (int az = 0; az < imageWidth; az++)
            {
                for (int bz = 0; bz < imageHeight; bz++)
                {
                    if (i < j)
                    {
                        if (abs(a-az) < 2 && abs(b-bz) < 2)
                        {
                            edge_descriptor edge_i_to_j, edge_j_to_i;
                            bool edge_i_to_j_exists = edge(verts[i],verts[j],g).second;

                            if (edge_i_to_j_exists == false)
                            {
                                bool bin1;
                                boost::tie(edge_i_to_j, bin1) = add_edge(verts[i], verts[j], g);
                            }
                            else
                            {
                                edge_i_to_j = edge(verts[i],verts[j],g).first;
                            }

                            bool edge_j_to_i_exists = edge(verts[j],verts[i],g).second;

                            if (edge_j_to_i_exists == false)
                            {
                                bool bin1;
                                boost::tie(edge_j_to_i, bin1) = add_edge(verts[j], verts[i], g);
                            }
                            else
                            {
                                edge_j_to_i =  edge(verts[j],verts[i],g).first;
                            }

                            reverse_edge[edge_i_to_j] = edge_j_to_i;
                            reverse_edge[edge_j_to_i] = edge_i_to_j;

                            double rs = Input[az][bz] - Input[a][b];
                            double rsa = exp(pow(rs, 2));
                            double rsb = (adjustedWeights[2] + adjustedWeights[3]*rsa)/8;
                            long vr = rsb;

                            capacity[edge_i_to_j] = capacity[edge_i_to_j] + 4*vr;
                            capacity[edge_j_to_i] = 0;

                            edge_descriptor edge_source_to_i = edge(verts[0],verts[i],g).first;
                            edge_descriptor edge_j_to_sink = edge(verts[j],verts[verticesCount - 1],g).first;

                            capacity[edge_source_to_i] = capacity[edge_source_to_i] + 2*vr;
                            capacity[edge_j_to_sink] = capacity[edge_j_to_sink] + 2*vr;
                        }
                    }
                    j++;
                }
            }
            i++;
        }
    }

    //Now find maximum flow:
    long flow;
    //flow = push_relabel_max_flow(g, src, sink);
    flow = edmonds_karp_max_flow(g, src, sink);


    qDebug() << " flow: " << flow;
    qDebug() << " verticesCount: " << verticesCount;

    //Now create new graph with same vertices as before, and edges as before only where capacity > 0 and residual = 0.
    typedef adjacency_list <vecS, vecS, undirectedS> UndirectedGraph;
    UndirectedGraph UndirectedG;

    graph_traits<Graph>::vertex_iterator u_iter, u_end;
    graph_traits<Graph>::out_edge_iterator ei, e_end;

    for (tie(u_iter, u_end) = vertices(g); u_iter != u_end; ++u_iter)
    {
        for (tie(ei, e_end) = out_edges(*u_iter, g); ei != e_end; ++ei)
        {
            int sor = *u_iter;
            int tar = target(*ei, g);

            if (capacity[*ei] > 0)
            {
                qDebug() << " Edge: " << sor << ", " << tar << " " << capacity[*ei] << " " << residual_capacity[*ei];
            }

            if (capacity[*ei] > 0 && residual_capacity[*ei] > 0)
            {
                if (capacity[*ei] - residual_capacity[*ei] > 0)
                {
                    add_edge(sor, tar, UndirectedG);
                    qDebug() << " Adding Undirected Edge: " << sor << " -> " << tar;
                }
            }
        }
    }
    //Add source to verticesCount connection.. so that it will always show up as a component
    add_edge(0, verticesCount, UndirectedG);

    //Add sink to verticesCount+1 connection.. so that it will always show up as a component
    add_edge(verticesCount-1, verticesCount+1, UndirectedG);

    //Find all the nodes connected to the sink-node here.. and set them to value 1.
    std::vector<int> component(num_vertices(UndirectedG));
    int num = connected_components(UndirectedG, &component[0]);

    qDebug() << "Connected components: " << num;
    qDebug() << "num_vertices: " << num_vertices(UndirectedG);
    if (num_vertices(UndirectedG) >= verticesCount - 1)
    {
        int sinkComponent = component[verticesCount - 1];
        qDebug() << "sinkComponent: " << sinkComponent;

        for (i = 0; i != component.size(); ++i)
        {
            qDebug() << " Vertex " << i << ": " << component[i];
            if (i > 0 && i < verticesCount - 1)
            {
                if (component[i] == sinkComponent)
                {
                    Ret[MapVertexToX[i-1]][MapVertexToY[i-1]] = 1;
                }
                else
                {
                    Ret[MapVertexToX[i-1]][MapVertexToY[i-1]] = 0;
                }
            }
            //cout << "Vertex " << i <<" is in component " << component[i] << endl;
        }
    }

    #ifdef PRINT_MAP_DEBUG_COMMENTS
    double eng = EvaluateOutputEnergy(Input, Ret);

    qDebug() << " Displaying Result...";
    qDebug() << " Output Energy: " << eng;
    for (size_t i = 0; i < imageWidth; i++)
    {
        for (size_t j = 0; j < imageHeight; j++)
        {
            qDebug() << " Input(" << i << ", " << j << ") = " << Input[i][j];
            qDebug() << " Result(" << i << ", " << j << ") = " << Ret[i][j];
        }
    }
    #endif

    return Ret;

}
